from ..exceptions import ModuleNotFoundException

class MongoWriter():
    def __init__(self, conn_str: str, database: str, collection: str) -> None:
        try:
            from pymongo import MongoClient
        except ImportError:
            raise ModuleNotFoundException("pymongo not installed. try `pip install pymongo`.")
        self.mdb = MongoClient(conn_str)
        self.collection = collection
        self.database = database

    def write_dataframe(self, df):
        records = df.to_dict()
        self.mdb[self.database][self.collection].insert_many(records)
        


from watchdog.events import PatternMatchingEventHandler
from watchdog.observers import Observer
from ..exceptions import FileSourcePathMissingException, FileExtensionNotSupportedException
import polars as pl
import time
from pathlib import Path
from .utils import _writer, READERS, SUPPORTED_FILE_WRITERS

class FileHandler(PatternMatchingEventHandler):
    def __init__(self, reader, transform, writer) -> None:
        self.reader = reader
        self.transform = transform
        self.writer = writer
        PatternMatchingEventHandler.__init__(self, patterns=[f'*.{self.reader.source}'], ignore_directories=True, case_sensitive=False)

    def on_created(self, event):
        #print(f"New {self.source} file created", event.src_path)
        try:
            df = READERS[self.reader.source](event.src_path)
            df = self.transform(df.lazy())
            if self.writer.source=='console':
                print(df.collect())
            elif self.writer.source in SUPPORTED_FILE_WRITERS:
                path = self.writer._options['path'].rstrip('/')
                filename = Path(event.src_path).stem
                fullpath = f"{path}/{filename}.{self.writer.source}"
                _writer(df.collect(), fullpath, self.writer.source)
        except Exception as e:
            print(e)

class FileProcessor():
    def __init__(self, reader, transform, writer) -> None:
        self.reader = reader
        self.transform = transform
        self.writer = writer

    def start(self):
        event_handler = FileHandler(self.reader, self.transform, self.writer)
        observer = Observer()
        if self.reader.file_source_path:
            path = self.reader.file_source_path
        elif 'path' in self.reader._options:
            path = self.reader._options['path']
        else:
            raise FileSourcePathMissingException("file source path not mentioned")
        observer.schedule(event_handler,path, recursive=True)
        observer.start()
        try:
            while True:
                time.sleep(1)
        except KeyboardInterrupt:
            observer.stop()
            observer.join()


from ..exceptions import ModuleNotFoundException

class ElasticSearch():
    def __init__(self, host: str, index: str, username: str =None, password: str =None, api_key: str = None) -> None:
        try:
            from elasticsearch import Elasticsearch
        except ImportError:
            raise ModuleNotFoundException("elasticsearch not installed. try `pip install elasticsearch`.")
        if username is not None and password is not None:
            self.es = ElasticSearch([host], basic_auth=(username, password))
        elif api_key is not None:
            self.es = ElasticSearch([host], api_key=api_key)
        else:
            self.es = ElasticSearch([host])
        self.index = index

    def write_dataframe(self, df, index):
        from elasticsearch.helpers import bulk
        records = df.to_dict()
        actions = [
            {
                "_index": self.index,
                "_source": doc
            }
            for doc in records
        ]
        bulk(self.es, actions)


from .core import StreamProcessor
from importlib import resources
try:
    import tomllib
except ModuleNotFoundError:
    import tomli as tomllib

# Version of the polars-streaming package
__version__ = "0.2.0"


from ..exceptions import FileExtensionNotSupportedException
import polars as pl

READERS = {'csv': pl.read_csv, 'parquet': pl.read_parquet, 'json': pl.read_json, 'avro': pl.read_avro}
SUPPORTED_FILE_WRITERS = ['csv','parquet','json','avro']

def _writer(df, path, extension):
    if extension == 'csv':
        df.write_csv(path)
    elif extension == 'parquet':
        df.write_parquet(path)
    elif extension == 'json':
        df.write_json(path)
    elif extension == 'avro':
        df.write_avro(path)
    else:
        raise FileExtensionNotSupportedException(f'UnSupported File Extension: {extension}')
    
def trigger_time_extracter(time_str):
    t, unit = time_str.split()
    if unit in ['seconds', 'second', 'sec', 'secs']:
        return int(t)
    elif unit in ['minutes', 'minute', 'min', 'mins']:
        return int(t)*60
    elif unit in ['hours', 'hour', 'hr', 'hrs']:
        return int(t)*60*60



import polars as pl
import schedule
from .utils import _writer, SUPPORTED_FILE_WRITERS, trigger_time_extracter
from ..exceptions import ModuleNotFoundException

options_mapper = {'kafka.bootstrap.servers':'bootstrap.servers',
                #'subscribe': 'subscribe',
                #'includeHeaders': 'includeHeaders',
                'startingOffsets': 'auto.offset.reset',
                #'endingOffsets': 'endingOffsets',
                #'subscribePattern': 'subscribePattern',
                'kafka.group.id': 'group.id'}

class KafkaProcessor():
    def __init__(self, reader, transform, writer):
        try:
            import confluent_kafka
        except ImportError:
            raise ModuleNotFoundException("confluent_kafka not installed. try `pip install confluent-kafka`.")
        self.reader = reader
        self.writer = writer
        self.transform = transform

    def _kafka_option_mapper(self):
        options = self.reader._options
        kafka_config = {}
        for key, value in options.items():
            if key in options_mapper:
                kafka_config[options_mapper[key]] = value
        self.kafka_config = kafka_config

    def apply_transform(self, transformation):
        df = pl.from_records(self.mini_batch)
        df = transformation(df.lazy())
        if self.writer.source=='console':
            print(f"Batch: {self.batch_count}")
            print(df.collect())
        elif self.writer.source in SUPPORTED_FILE_WRITERS:
            path = self.writer._options['path'].rstrip('/')
            filename = f"batch_{self.batch_count}"
            fullpath = f"{path}/{filename}.{self.writer.source}"
            _writer(df.collect(), fullpath, self.writer.source)
        self.batch_count+=1
        self.mini_batch.clear()

    def start(self):
        from confluent_kafka import Consumer, KafkaException
        self._kafka_option_mapper()
        consumer = Consumer(self.kafka_config)
        topics = self.reader._options['subscribe'].split(',')
        consumer.subscribe(topics)
        self.mini_batch = []
        ptime = trigger_time_extracter(self.writer.processing_time)
        schedule.every(ptime).seconds.do(self.apply_transform, transformation = self.transform)
        self.batch_count = 0
        try:
            while True:
                event = consumer.poll(1.0)
                if event is None:
                    continue
                if event.error():
                    raise KafkaException(event.error())
                else:
                    val = event.value().decode('utf8')
                    consumer.commit(event)
                self.mini_batch.append(val)
                schedule.run_pending()
        except KeyboardInterrupt:
            print('Canceled by user.')
        finally:
            consumer.close()


class FileSourcePathMissingException(Exception):
    def __init__(self,message):
        self.message = message

class FileExtensionNotSupportedException(Exception):
    def __init__(self,message):
        self.message = message

class NotImplementedError(Exception):
    def __init__(self,message):
        self.message = message

class ModuleNotFoundException(Exception):
    def __init__(self,message):
        self.message = message


import polars as pl
import socket
import schedule
from .utils import _writer, SUPPORTED_FILE_WRITERS, trigger_time_extracter

class SocketProcessor():
    def __init__(self, reader, transform, writer):
        self.reader = reader
        self.writer = writer
        self.transform = transform

    def apply_transform(self, transformation):
        df = pl.from_records(self.mini_batch)
        df = transformation(df.lazy())
        if self.writer.source=='console':
            print(f"Batch: {self.batch_count}")
            print(df.collect())
        elif self.writer.source in SUPPORTED_FILE_WRITERS:
            path = self.writer._options['path'].rstrip('/')
            filename = f"batch_{self.batch_count}"
            fullpath = f"{path}/{filename}.{self.writer.source}"
            _writer(df.collect(), fullpath, self.writer.source)
        self.batch_count+=1
        self.mini_batch.clear()

    def start(self):
        host = self.reader._options['host']
        port = self.reader._options['port']
        if 'bufferSize' in self.reader._options:
            buffer = self.reader._options['bufferSize']
        else:
            buffer = 1024
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.connect((host, port))
        self.mini_batch = []
        ptime = trigger_time_extracter(self.writer.processing_time)
        schedule.every(ptime).seconds.do(self.apply_transform, transformation = self.transform)
        self.batch_count = 0
        try:
            while True:
                msg = s.recv(buffer)
                msg = msg.decode("utf-8")
                self.mini_batch.append(msg)
                schedule.run_pending()
        except KeyboardInterrupt:
            print('Canceled by user.')
        finally:
            s.close()


import polars as pl
from glob import glob
import os

FILE_SOURCES = ['csv','parquet','json','avro']

READERS = {'csv': pl.read_csv, 'parquet': pl.read_parquet, 'json': pl.read_json, 'avro': pl.read_avro}

class DataStreamReader():

    def __init__(self):
        self._options = {}

    def format(self, source):
        self.source = source
        return self

    def option(self, key: str, value):
        self._options[key] = value
        return self

    def options(self, options):
        for key, value in options.items():
            self._options[key] = value
        return self

    def schema(self, schema):
        self.schema = schema

    def load(self, path: str = None, preFetchFirstBatch: str = False):
        self.file_source_path = path
        if preFetchFirstBatch:
            if self.source in FILE_SOURCES:
                if self.file_source_path is not None:
                    files = glob(f"{self.file_source_path.rstrip('/')}/*.{self.source}")
                    files.sort(key=os.path.getmtime)
                    self.df = READERS[self.source](files[0])
        return self
        
class DataStreamWriter():
    def __init__(self):
        self._options = {}

    def format(self, source):
        self.source = source
        return self

    def option(self, key, value):
        self._options[key] = value
        return self
    
    def options(self, options):
        for key, value in options.items():
            self._options[key] = value
        return self
    
    def outputMode(self, mode):
        self.output_mode = mode
        return self

    def trigger(self, ProcessingTime="1 second"):
        self.processing_time = ProcessingTime
        return self



from .processors.fileProcessor import FileProcessor
from .readwriter import DataStreamReader, DataStreamWriter
from .processors.kafkaProcessor import KafkaProcessor
from .processors.socketProcessor import SocketProcessor
from .exceptions import NotImplementedError

FILE_SOURCES = ['csv','parquet','json','avro']

class StreamProcessor():

    @property
    def readStream(self) -> DataStreamReader:
        self.reader = DataStreamReader()
        return self.reader

    @property
    def writeStream(self) -> DataStreamWriter:
        self.writer = DataStreamWriter()
        return self.writer
    
    def preFetchedDF(self):
        return self.reader.df

    def add_transform(self, transforms):
        self.transform = transforms

    def start(self):
        if self.reader.source in FILE_SOURCES:
            FileProcessor(self.reader, self.transform, self.writer).start()
        elif self.reader.source == 'kafka':
            KafkaProcessor(self.reader, self.transform, self.writer).start()
        elif self.reader.source == 'socket':
            SocketProcessor(self.reader, self.transform, self.writer).start()
        else:
            raise NotImplementedError(f'Not Implemented ReadStream Source: {self.reader.source}')
